
import argparse
import numpy as np
from hyperopt import hp
import torch

def parameters(path, debug):

    # args=DefaultParameters(path)
    args={}
    args['path'] = path

    # Use optimized parameters depending on the experiment.
    if 'STK11' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # Architecture of the CNN. Choose between resnet50 or resnet18. resnet50 is slow but is capable of learning more complex patterns. resnet18 is faster but learns simpler patterns. So far we have used resnet50.
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate used to train the CNN. Do not change this value.
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay used to train the CNN. Do not change this value.
        args['PCL_Temperature']= 0.07 # Softmax temperature used to train the CNN. Do not change this value.
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # Specify the GPU where the training is carried out.
        args['PCL_Epochs']=400 # Number of epochs to run PCL. Increase the numnber so that PCL achieves at least a 85% of top1 accuracy.
        args['PCL_N_Workers']=1 # Number of workers used to parallelize data loading. A higher number of workers will result in a faster execution but will require more RAM memory usage. 
        args['PCL_N_Crops_per_Image']=300 # Number of image crops used to train the model in each iteration. Higher number of crops will learn more features from patches from one image but requires more RAM and GPU memory usage.
        args['PCL_Patch_Size']=30 # Size of image patch (in pixels). We saw best performance when the patch embeds 1 to 3 cells.
        args['PCL_Stride']=0 # Size of stride between consecutive patches. Do not use it. It is not yet implemented. :(
        args['PCL_Alpha_L']=1.15 # Size ratio between image crop and image patch. Check paper for better explanation.
        args['PCL_Z_Score']= True # Whether to apply z_score per channel/marker. Seen better performance when using it. 
        args['PCL_Batch_Size']= 4 # Number of images used per iteration. A higher batch size results in learning more features and doing it in a more stable way but will require more RAM and GPU memory usage.    
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of the output image patch. A higher value results in more learned features but requires more RAM and GPU memory usage.
        args['PCL_eliminate_Black_Background'] = True # Whether to eliminate from the training the black background. Use it for Multiplex fluorescence images.
        args['PCL_eliminate_White_Background'] = False # Whether to eliminate from the training the whit background. Use it for H&E white-field microscopy images.        
        args['PCL_Color_perturbation_Augmentation'] = False # Whether to use color normalization augmentation in H&E white-field microscopy images. Used to force PCL to learn morphological features and avoid learning color staining features. 

        # Name of the columnn from Image_Labels.xslx used to differentiate subject types.
        args['experiment_Label'] = ['Unsupervised'] 

        # General parameters used for Machine Learning
        args['NaroNet_n_workers'] = 1 # Number of workers used to parallelize data loading. A higher number of workers will result in a faster execution but will require more RAM memory usage. 
        args['epochs'] = 3 # Number of epochs to run PCL. Increase the number so that all function losses find a plateau without overfitting
        args['weight_decay'] = 0 # Weight Decay used to train the CNN. Do not change this value.
        args['batch_size'] = 12 # Number of subjects used per iteration. A higher batch size results in learning more features and doing it in a more stable way but will require more RAM and GPU memory usage.    
        args['lr'] = 0.001 # Learning Rate used to train the CNN. Do not change this value.
        args['dropoutRate'] = 0 # Dropout rate 
        args['training_MODE'] = 'TrainALL' # Train mode selection. 'TrainALL' uses 100% of the data to train (no validation). 'CrossValidation' divides data in k-folds to obtain generability of the model.
        args['folds'] = 1 # Number of folds used in the cross-validation training/testing mode. Check https://machinelearningmastery.com/k-fold-cross-validation/ for more info. 
        args['device'] = 'cuda:0' # Specify the GPU where the training is carried out.
        args['Batch_Normalization'] = True # Whether to apply batch normalization just before patch assignment. Do not change this value. If False it will not work.
        args['dataAugmentationPerc'] = 0 # Artificially augment subjects in each iteration. Value specifies percentage of nodes of the graph changed where 1 stands for 100% and 0 implies no data augmentation used.
            
        # NaroNet specific parameters
        args['N_Phenotypes'] = 8 # Number of local phenotypes. Use low values when you want to find only 2 or 3 relevant phenotypes. Use larger values to find multiple relevant phenotypes.
        args['N_Neighborhoods'] = 12 # Number of cellular neighborhoods. Use low values when you want to find only 2 or 3 relevant neighborhoods. Use larger values to find multiple relevant neighborhoods.
        args['N_Areas'] = 4 # Number of tissue areas. Use low values when you want to find only 2 or 3 ares. Use larger values to find multiple relevant areas.     
        args['n_hops_neighborhoods'] = 1 # Number of hops used to learn neighborhoods using GNNs. Do not change this value.
        args['Phenotype_Learning'] = True # Whether to learn phenotypes or not  
        args['Neighborhood_Learning'] = True # Whether to learn neighborhoods or not
        args['Area_Learning'] = True # Whether to learn areas or not        
        args['attntnThreshold'] = 0 # Attention threshold for patch assignment. Patches with a confidence value below this are not considered for subject prediction
        args['1Patch1Cluster'] = False # Whether to let patches be from several TMEs at the same time or not.
        
        # Losses
        args['ContrastACC_Pheno'] = 1000 # Relevance of contrast accuracy loss for phenotypes
        args['ContrastACC_Neigh'] = 1000 # Relevance of contrast accuracy loss for neighborhoods
        args['ContrastACC_Area'] = 0 # Relevance of contrast accuracy loss for areas
        args['PatchEntropy_Pheno'] = 0 # Relevance of contrast accuracy loss for areas
        args['PatchEntropy_Neigh'] = 0#1 if debug=='Index' else hp.choice("PatchEntropy_Neigh", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.1
        args['PatchEntropy_Area'] = 1#4 if debug=='Index' else hp.choice("PatchEntropy_Area", [1,0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.0001                                             

        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False
        args['Perc_high_conf_patches'] = 0.0001 # 1 is 100%. Great values generates phenotypes and neighborhoods with many patches. A low values generates rare populations.                                  
        args['TSNE_Perc_Pat'] = 0.5 # Percentage of patients to use to generate the tsne.     


    elif 'Melanoma_ProgressionvsBaseline' in path:            
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=2 # Number of data loading workers 
        args['PCL_Epochs']=300 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=300 # Number of crops chosen per image
        args['PCL_Patch_Size']=50 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker 
        args['PCL_Batch_Size']= 6 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature                
        args['PCL_eliminate_Black_Background'] = False
        args['PCL_eliminate_White_Background'] = True   
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 
        args['PCL_Color_perturbation_Augmentation'] = True # Color augmentation for H&E tissue staining.

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Sample_Type'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 700  

        # Optimization Parameters
        args['NaroNet_n_workers'] = 1 # Number of data loading workers
        args['epochs'] = 400# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/3),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 0#3 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.0001
        args['batch_size'] = 4 if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 16, 32]) if debug=='Object' else 32
        args['lr'] = 2 if debug=='Index' else hp.choice('lr', [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 0.01

        # General        
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.            
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 1 if debug=='Index' else hp.choice('clusters1',[5,8,15,20]) if debug=='Object' else 8     
        args['N_Neighborhoods'] = 2 if debug=='Index' else hp.choice('clusters2',[6,11,12,21]) if debug=='Object' else 12
        args['N_Areas'] = 1 if debug=='Index' else hp.choice('clusters3',[2,6,7]) if debug=='Object' else 6      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True  
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = True# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  

        # Losses 
        args['ContrastACC_Pheno'] = 100#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 100#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 100#.0001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001            
        args['Perc_high_conf_patches'] = 0.001 # 1 is 100%. Great values generates phenotypes and neighborhoods with many patches. A low values generates rare populations.                                  
        args['PatchEntropy_Pheno'] = 0#.01#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#.01#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#.0001#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1                                  

        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False
        args['TSNE_Perc_Pat'] = 0.33 # Percentage of patients to use to generate the tsne. 


    elif 'LF_Recurrence' in path:            
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=8 # Number of data loading workers 
        args['PCL_Epochs']=400 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=70 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.1 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker 
        args['PCL_Batch_Size']= 8 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature    
        args['PCL_eliminate_Black_Background'] = True
        args['PCL_eliminate_White_Background'] = False
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 
        args['PCL_Color_perturbation_Augmentation'] = False # Color augmentation for H&E tissue staining.

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Relapse'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 700  

        # Optimization Parameters
        args['NaroNet_n_workers'] = 1
        args['epochs'] = 200# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 309#max(int(args['epochs']/3),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 0#3 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.0001
        args['batch_size'] = 4 if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 12, 28]) if debug=='Object' else 28
        args['lr'] = 1 if debug=='Index' else hp.choice('lr', [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 0.1

        # General        
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.            
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 1 if debug=='Index' else hp.choice('clusters1',[5,8,15,20]) if debug=='Object' else 8     
        args['N_Neighborhoods'] = 2 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 16 
        args['N_Areas'] = 1 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 4      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = False  
        args['Area_Learning'] = False
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = False# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  

        # Losses 
        args['ContrastACC_Pheno'] = 1000#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 0#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#.0001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['PatchEntropy_Pheno'] = 0#.01#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#.01#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#.0001#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1                                  

        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False
        args['TSNE_Perc_Pat'] = 0.5 # Percentage of patients to use to generate the tsne. 


    elif 'LF_19H' in path:            
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=0 # Number of data loading workers 
        args['PCL_Epochs']=400 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=200 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker 
        args['PCL_Batch_Size']= 6 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature         
        args['PCL_eliminate_Black_Background'] = True
        args['PCL_eliminate_White_Background'] = False       
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 
        args['PCL_Color_perturbation_Augmentation'] = False # Color augmentation for H&E tissue staining.

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Relapse'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 700  

        # Optimization Parameters
        args['epochs'] = 8# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/3),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 0#3 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0.0001
        args['batch_size'] = 26# if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 16, 22]) if debug=='Object' else 22
        args['lr'] = 0.1#0 if debug=='Index' else hp.choice('lr', [1,0.1,0.01]) if debug=='Object' else 1

        # General        
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.            
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 1 if debug=='Index' else hp.choice('clusters1',[5,8,15,20]) if debug=='Object' else 8     
        args['N_Neighborhoods'] = 2 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 16 
        args['N_Areas'] = 1 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 4      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True  
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = False# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  


        # Losses
        args['ContrastACC_Pheno'] = 1#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 1#.001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 1#.0001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['PatchEntropy_Pheno'] = 0#.01#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#.01#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#.0001#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1                                  

        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False

    elif 'Urothelial_ClinBen' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=8 # Number of data loading workers 
        args['PCL_Epochs']=80 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=40 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 8 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature    
        args['PCL_eliminate_Black_Background'] = True 
        args['PCL_eliminate_White_Background'] = False            
        args['PCL_Color_perturbation_Augmentation'] = False # Whether to use color normalization augmentation in H&E white-field microscopy images. Used to force PCL to learn morphological features and avoid learning color staining features. 
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Clinical_Benefit'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['NaroNet_n_workers'] = 1
        args['epochs'] = 1000# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/5),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 4 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0
        args['batch_size'] = 3 if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 24]) if debug=='Object' else 24
        args['lr'] = 3 if debug=='Index' else hp.choice('lr', [1,0.1,0.01,0.001,0.001]) if debug=='Object' else 0.001

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 6#2 if debug=='Index' else hp.choice('clusters1',[5,10,15,20]) if debug=='Object' else 15     
        args['N_Neighborhoods'] = 8#1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 4#0 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 2      
        args['n_hops_neighborhoods'] = 2 
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = False
        args['Area_Learning'] = False
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = False# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
                
        # Losses
        args['ContrastACC_Pheno'] = 1# 4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 0 # 4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001 
        args['Perc_high_conf_patches'] = 0.1 # 1 is 100%. Great values generates phenotypes and neighborhoods with many patches. Low values generate rare populations.
        args['PatchEntropy_Pheno'] = 0#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = True
        args['TSNE_Perc_Pat'] = 0 # Percentage of patients used to calculate TSNE. 1 is 100% of patients
        args['Pheno_Perc_Pat'] = 1 # Percentage of patients used to calculate TSNE. 1 is 100% of patients
        args['Pheno_Excluded_from_analysis'] = [] # Percentage of patients used to calculate TSNE. 1 is 100% of patients
        args['Neigh_Excluded_from_analysis'] = [] # Percentage of patients used to calculate TSNE. 1 is 100% of patients


    elif 'Endometrial_Survival' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=8 # Number of data loading workers 
        args['PCL_Epochs']=250 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=30 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 40 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature    
        args['PCL_eliminate_Black_Background'] = True 
        args['PCL_eliminate_White_Background'] = False        
        args['PCL_Color_perturbation_Augmentation'] = False # Whether to use color normalization augmentation in H&E white-field microscopy images. Used to force PCL to learn morphological features and avoid learning color staining features. 
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['POLE Mutation'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['NaroNet_n_workers'] = 1
        args['epochs'] = 75# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/5),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 4 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0
        args['batch_size'] = 3 if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 64]) if debug=='Object' else 64
        args['lr'] = 1 if debug=='Index' else hp.choice('lr', [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 0.001

        # General
        args['training_MODE'] = 'CrossValidation' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 12
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 6#2 if debug=='Index' else hp.choice('clusters1',[5,10,15,20]) if debug=='Object' else 15     
        args['N_Neighborhoods'] = 8#1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 4#0 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 2      
        args['n_hops_neighborhoods'] = 3
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = False# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
                
        # Losses
        args['ContrastACC_Pheno'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001 
        args['Perc_high_conf_patches'] = 0.0001 # 1 is 100%. Great values generates phenotypes and neighborhoods with many patches. Low values generate rare populations.
        args['PatchEntropy_Pheno'] = 0.001#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0.001#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0.001#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = True
        args['TSNE_Perc_Pat'] = 1 # Percentage of patients to use to generate the tsne. 

    elif 'Tumor_subtype' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=8 # Number of data loading workers 
        args['PCL_Epochs']=80 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=40 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 8 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature                
        args['PCL_eliminate_Black_Background'] = False
        args['PCL_eliminate_White_Background'] = False
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Clinical_Benefit'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['epochs'] = 500
        args['lr_decay_factor'] = 0.1
        args['lr_decay_step_size'] = 10000
        args['weight_decay'] = 0
        args['batch_size'] = 24
        args['lr'] = 1 

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 8#2 if debug=='Index' else hp.choice('clusters1',[5,10,15,20]) if debug=='Object' else 15     
        args['N_Neighborhoods'] = 16#1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 4#0 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 2      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
                
        # Losses
        args['ContrastACC_Pheno'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['PatchEntropy_Pheno'] = 0#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False


    elif 'Mary_Helen' in path or 'MH_test' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=8 # Number of data loading workers 
        args['PCL_Epochs']=200 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=40 # Number of crops chosen per image
        args['PCL_Patch_Size']=40 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 20 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature    
        args['PCL_eliminate_Black_Background'] = True 
        args['PCL_eliminate_White_Background'] = False        
        args['PCL_Color_perturbation_Augmentation'] = False # Whether to use color normalization augmentation in H&E white-field microscopy images. Used to force PCL to learn morphological features and avoid learning color staining features. 
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Treatment_Administered'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['NaroNet_n_workers'] = 1
        args['epochs'] = 1000# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/5),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 0
        args['batch_size'] = 24
        args['lr'] = 0.001

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 8
        args['N_Neighborhoods'] = 12
        args['N_Areas'] = 4
        args['n_hops_neighborhoods'] = 1
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = False
        args['dropoutRate'] = 0
        args['attntnThreshold'] = 0
        args['1Patch1Cluster'] = False
                
        # Losses
        args['ContrastACC_Pheno'] = 1000#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 1000#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001 
        args['Perc_high_conf_patches'] = 0.0001 # 1 is 100%. Great values generates phenotypes and neighborhoods with many patches. Low values generate rare populations.
        args['PatchEntropy_Pheno'] = 0#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = True
        args['TSNE_Perc_Pat'] = 0 # Percentage of patients to use to generate the tsne. 
        args['Pheno_Perc_Pat'] = 0.05 # Percentage of patients to use to generate the tsne. 


    elif 'Melanoma_ProgressionvsBaseline' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=1 # Number of data loading workers 
        args['PCL_Epochs']=300 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=200 # Number of crops chosen per image
        args['PCL_Patch_Size']=50 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 4 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature                
        args['PCL_eliminate_Black_Background'] = False
        args['PCL_eliminate_White_Background'] = True
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 
        args['PCL_Color_perturbation_Augmentation'] = True # Color augmentation for H&E tissue staining.

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Sample_Type'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['NaroNet_n_workers'] = 1
        args['epochs'] = 1000
        args['lr_decay_factor'] = 0.1
        args['lr_decay_step_size'] = 10000
        args['weight_decay'] = 0
        args['batch_size'] = 24
        args['lr'] = 0.001 

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 6#2 if debug=='Index' else hp.choice('clusters1',[5,10,15,20]) if debug=='Object' else 15     
        args['N_Neighborhoods'] = 8#1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 4#0 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 2      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = False
                
        # Losses
        args['ContrastACC_Pheno'] = 100#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 100#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['Perc_high_conf_patches'] = 0.001 # 1 is 100%. Great values generates phenotypes and neighborhoods with many patches. Low values generate rare populations.
        args['PatchEntropy_Pheno'] = 0#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = False
        args['TSNE_Perc_Pat'] = 0.5 # Percentage of patients used to calculate TSNE. 1 is 100% of patients


    elif 'AKOYA_Arlene_Tonsil' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=10 # Number of data loading workers 
        args['PCL_Epochs']=100 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=20 # Number of crops chosen per image
        args['PCL_Patch_Size']=40 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 50 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature                
        args['PCL_eliminate_Black_Background'] = False
        args['PCL_eliminate_White_Background'] = False
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 
        args['PCL_Color_perturbation_Augmentation'] = False # Color augmentation for H&E tissue staining.

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Unsupervised'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['NaroNet_n_workers'] = 1
        args['epochs'] = 1000
        args['lr_decay_factor'] = 0.1
        args['lr_decay_step_size'] = 10000
        args['weight_decay'] = 0
        args['batch_size'] = 128
        args['lr'] = 0.1 

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 72#2 if debug=='Index' else hp.choice('clusters1',[5,10,15,20]) if debug=='Object' else 15     
        args['N_Neighborhoods'] = 36#1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 6#0 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 2      
        args['n_hops_neighborhoods'] = 1
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = False
                
        # Losses
        args['ContrastACC_Pheno'] = 1#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 1#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001     
        args['Perc_high_conf_patches'] = 0.0001 # 1 is 100%. Great values generates phenotypes and neighborhoods with many patches. Low values generate rare populations.
        args['PatchEntropy_Pheno'] = 0#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 1#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = True
        args['TSNE_Perc_Pat'] = 0 # Percentage of patients used to calculate TSNE. 1 is 100% of patients
        args['Pheno_Perc_Pat'] = 1 # Percentage of patients used to calculate TSNE. 1 is 100% of patients
        args['Pheno_Excluded_from_analysis'] = [25] # Percentage of patients used to calculate TSNE. 1 is 100% of patients
        args['Neigh_Excluded_from_analysis'] = [] # Percentage of patients used to calculate TSNE. 1 is 100% of patients
    
    elif 'Pancreas_Adenocarcinoma' in path:            
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=10 # Number of data loading workers 
        args['PCL_Epochs']=600 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=40 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 20 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.0003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature                
        args['PCL_eliminate_Black_Background'] = False
        args['PCL_eliminate_White_Background'] = True
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 
        args['PCL_Color_perturbation_Augmentation'] = False # Color augmentation for H&E tissue staining.

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Patient_Status'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['NaroNet_n_workers'] = 1
        args['epochs'] = 1000
        args['lr_decay_factor'] = 0.1
        args['lr_decay_step_size'] = 10000
        args['weight_decay'] = 0
        args['batch_size'] = 24
        args['lr'] = 0.0001 

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 8
        args['N_Neighborhoods'] = 12
        args['N_Areas'] = 4
        args['n_hops_neighborhoods'] = 2
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = False
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
        args['1Patch1Cluster'] = False
                
        # Losses
        args['ContrastACC_Pheno'] = 1000#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 1000#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001     
        args['Perc_high_conf_patches'] = 0.0001 # 1 is 100%. Great values generates phenotypes and neighborhoods with many patches. Low values generate rare populations.
        args['PatchEntropy_Pheno'] = 0#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = True
        args['TSNE_Perc_Pat'] = 0 # Percentage of patients used to calculate TSNE. 1 is 100% of patients
        args['Pheno_Perc_Pat'] = 1 # Percentage of patients used to calculate TSNE. 1 is 100% of patients
        args['Pheno_Excluded_from_analysis'] = [] # Percentage of patients used to calculate TSNE. 1 is 100% of patients
        args['Neigh_Excluded_from_analysis'] = [] # Percentage of patients used to calculate TSNE. 1 is 100% of patients


    else:
        
        # Patch contrastive learning parameters
        args['PCL_CNN_Architecture']='resnet50' # or resnet18
        args['PCL_N_Workers']=8 # Number of data loading workers 
        args['PCL_Epochs']=80 # Number of total epoch to run
        args['PCL_N_Crops_per_Image']=40 # Number of crops chosen per image
        args['PCL_Patch_Size']=30 # Size of patches in pixels
        args['PCL_Stride']=0 # Size of stride between consecutive patches
        args['PCL_Alpha_L']=1.15 # size ratio between image crop and image patch
        args['PCL_Z_Score']= True # Whether to apply z score per channel/marker
        args['PCL_Batch_Size']= 1 # Number of images per iteration
        args['PCL_Weight_Decay']= 1e-4 # Weight Decay
        args['PCL_Out_Dimensions']= 128 # Feature dimensions of patch
        args['PCL_Learning_Rate']= 0.003 # Learning Rate
        args['PCL_Temperature']= 0.07 # Softmax temperature    
        args['PCL_eliminate_Black_Background'] = False
        args['PCL_eliminate_White_Background'] = False            
        args['PCL_GPU_INDEX']= torch.device('cuda:0') # GPU where the training is carried out 

        # Label you want to infer with respect the images.
        args['experiment_Label'] = ['Clinical_Benefit'] 

        # Architecture Search parameters
        args['num_samples_architecture_search'] = 1500  

        # Optimization Parameters
        args['epochs'] = 1000# if debug else hp.quniform('epochs', 5, 25, 1)
        args['lr_decay_factor'] = 0.1# if debug else hp.uniform('lr_decay_factor', 0, 0.75)
        args['lr_decay_step_size'] = 10000#max(int(args['epochs']/5),1)# if debug else hp.quniform('lr_decay_step_size', 2, 20, 1)        
        args['weight_decay'] = 4 if debug=='Index' else hp.choice('weight_decay',[0.1,0.01,0.001,0.0001,0]) if debug=='Object' else 0
        args['batch_size'] = 3 if debug=='Index' else hp.choice('batch_size', [2, 4, 8, 24]) if debug=='Object' else 24
        args['lr'] = 0 if debug=='Index' else hp.choice('lr', [1,0.1,0.01,0.001,0.0001]) if debug=='Object' else 1

        # General
        args['training_MODE'] = 'TrainALL' # TrainALL, NestedCrossValidation, CrossValidation # Whether to use nested cross validation or cross validation.        
        args['folds'] = 1
        args['device'] = 'cuda:0'
        args['Batch_Normalization'] = True
        args['dataAugmentationPerc'] = 0 #0 if debug=='Index' else hp.choice("dataAugmentationPerc", [0,0.0001,0.001]) if debug=='Object' else 0  

        # Neural Network
        args['N_Phenotypes'] = 8#2 if debug=='Index' else hp.choice('clusters1',[5,10,15,20]) if debug=='Object' else 15     
        args['N_Neighborhoods'] = 16#1 if debug=='Index' else hp.choice('clusters2',[6,11,16,21]) if debug=='Object' else 11 
        args['N_Areas'] = 4#0 if debug=='Index' else hp.choice('clusters3',[2,4,7]) if debug=='Object' else 2      
        args['Phenotype_Learning'] = True  
        args['Neighborhood_Learning'] = True
        args['Area_Learning'] = True
        args['dropoutRate'] = 0# 1 if debug=='Index' else hp.choice('dropoutRate', [0.1, 0.2, 0.3,0.4]) if debug=='Object' else 0.2        
        args['attntnThreshold'] = 0# if debug=='Index' else hp.choice('attntnThreshold', [0,.2,.4,.6,.8]) if debug=='Object' else 0  
                
        # Losses
        args['ContrastACC_Pheno'] = 0.01#4 if debug=='Index' else hp.choice("orthoColor_Lambda0", [1,0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                
        args['ContrastACC_Neigh'] = 0.01#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['ContrastACC_Area'] = 0.0001#4 if debug=='Index' else hp.choice("orthoColor_Lambda1", [1, 0.1,0.01,0.001,0.0001,0.00001]) if debug=='Object' else 0.0001                                              
        args['PatchEntropy_Pheno'] = 0.01#3 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda0", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.001   
        args['PatchEntropy_Neigh'] = 0.01#4 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda1", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0
        args['PatchEntropy_Area'] = 0.0001#1 if debug=='Index' else hp.choice("min_Cell_entropy_Lambda2", [1,0.1,0.01,0.001,0]) if debug=='Object' else 0.1          
        
        # Bioinsights parameters
        args['Bio_save_Orig_im'] = True


    return args


import os
import shutil

import torch
import yaml


def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, 'model_best.pth.tar')


def save_config_file(model_checkpoints_folder, args):
    if not os.path.exists(model_checkpoints_folder):
        os.makedirs(model_checkpoints_folder)
        with open(os.path.join(model_checkpoints_folder, 'config.yml'), 'w') as outfile:
            yaml.dump(args, outfile, default_flow_style=False)


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(1 / batch_size))
        return res



import os 
import numpy as np
import copy
import matplotlib.pyplot as plt
import statistics as st

def ObtainMultivariateIntersectInSynthetic(dataset, statisticalTests, clusters, IndexAndClass, num_classes, attentionLayer,isTraining):
    # statisticalTest: List of columns with statistical difference. 1.p-value, 2.Cluster step, 3. column of the heatmap        
    stsTest=statisticalTests[0]
    IntersecIndex=[]
    for nClust, clusterToProcess in enumerate(clusters):      
        IntersecIndex.append([])
        if isTraining and (nClust==0 or nClust==2):
            continue           
           
        # Create directory for this set of images.        
        thisFolder = dataset.bioInsights_dir+'/InterpretabilityQuant/BlueIsCluster_GreenIsGroundTruth_ClusterLevel_{}'.format(clusters[nClust])
        if not os.path.exists(thisFolder):
            os.makedirs(thisFolder)

        # For each image.
        for count, subject_info in enumerate(IndexAndClass): 
            statisticalTest =  statisticalTests[count]
            
            # Apply mask to patch
            for patchIDX in range(dataset.findLastIndex(subject_info[1])+1):                       
                
                # Pixel-to-cluster Matrix
                clust0 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], patchIDX, clusters[0]))                 
                if nClust>0:
                    clust1 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], patchIDX, clusters[1]))                                    
                if nClust>1:
                    clust2 = np.load(dataset.processed_dir_cell_types+'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(subject_info[1], clusters[2]))                    
                    clust2 = np.matmul(clust1,clust2)
                clust = clust0 
                clust = clust1 if nClust==1 else clust
                clust = clust2 if nClust==2 else clust                
                
                # Load patch canvas
                Patch_im = np.zeros(np.load(dataset.root+'Raw_Data/Images/'+subject_info[0]+'.npy').shape[:2])

                # Create Patch Label image using the patch size.                         
                division = np.floor(Patch_im.shape[0]/dataset.patch_size)
                lins = np.repeat(list(range(int(division))), dataset.patch_size)
                lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
                for y_indx, y in enumerate(range(int(division))):
                    Patch_im[:int(division*dataset.patch_size),y_indx*dataset.patch_size:(y_indx+1)*dataset.patch_size] = lins+int(y_indx*division)
                Patch_im = Patch_im.astype(int)
                if ('V_H' in dataset.root) or ('V3' in dataset.root) or ('V4' in dataset.root) or ('V2' in dataset.root) or ('V1' in dataset.root):
                    Patch_im = np.transpose(Patch_im)

                #
                cell_type_top1 = np.apply_along_axis(np.argmax, axis=1, arr=clust)                        
                
                # Initialize values
                Thresholds=[0,50,75,95]
                bestintersection=[-1,-1]
                IntersecIndex[nClust].append(0)
                for threshold in Thresholds:
                    AllClusters = copy.deepcopy(cell_type_top1)*0
                    
                    # Iterate through all significant clusters from this ClusterLevel
                    if 'Synthetic' in dataset.root: 
                        if int(statisticalTest[2][0])==clust.shape[1]:
                            AllClusters += statisticalTest[2][1]==cell_type_top1
                            AllClusters[clust[:,statisticalTest[2][1]]<threshold/100]=0
                    Patch_im=AllClusters[Patch_im]

                    # Load Ground-Truth
                    if ('SyntheticV2' in dataset.root) or ('SyntheticV_H2' in dataset.root): # fenotipo 4 y 5 de la region 2
                        Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==2
                    elif ('SyntheticV1' in dataset.root) or ('SyntheticV_H1' in dataset.root): # fenotipo 6 de la region 3
                        Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==3
                    elif ('SyntheticV3' in dataset.root) or ('SyntheticV_H3' in dataset.root):
                        Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==3
                    elif ('SyntheticV4' in dataset.root) or ('SyntheticV_H4' in dataset.root):
                        Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')
                        Ground_Truth = np.logical_or(Ground_Truth==2,Ground_Truth==3)
                    
                    # Calculate Intersection Index
                    intersection = np.logical_and(Patch_im, Ground_Truth)                    
                    if intersection.sum():
                        result = intersection.sum() / (1e-16+float(Patch_im.sum()))
                        bestintersection.append(result)
                                                
                        # In case its the best threshold so far.
                        if result>max(bestintersection[:-1]):
                            IntersecIndex[nClust][count] = max(bestintersection)
                            # Save GT and Image
                            RGBImage = np.zeros((Ground_Truth.shape[0],Ground_Truth.shape[1],3))
                            RGBImage[:,:,1] = Ground_Truth
                            RGBImage[:,:,2] = Patch_im
                            if not isTraining:
                                plt.imsave(thisFolder+'/IntersectIdx{}_Slide{}_Patch{}.png'.format(intersection.sum()/float(Patch_im.sum()),subject_info[1],patchIDX), RGBImage) 

                    else:
                        bestintersection.append(0)
                        IntersecIndex[nClust][count] = max(bestintersection)
               
        # Save Statistics Information
        if len(IntersecIndex[-1])==0:
            IntersecIndex[-1]=[-1]
        result_int = np.array(IntersecIndex[-1])
        eval_info = {'Intersect-Index': str(sum(result_int[result_int>0])/len(result_int[result_int>0]))+'+-'+str(st.pstdev(result_int[result_int>0])), ' Stats':statisticalTest}

        # Save epoch information into a log file.
        fn = thisFolder+"/Statistics.txt"                            
        print(str(fn))
        with open(fn, "w") as myfile:
            myfile.write(str(eval_info)+"\n")

    # Save epoch information into a log file.
    Int_Index = np.array(IntersecIndex).max(0)
    eval_info = {'Intersect-Index': str(np.mean(Int_Index[Int_Index>0]))+'+-'+str(np.std(Int_Index[Int_Index>0])), ' Stats':statisticalTest}
    fn = dataset.bioInsights_dir+'InterpretabilityQuant/Interpretability_Acuracy.txt'                
    print(str(fn))
    with open(fn, "w") as myfile:
        myfile.write(str(eval_info)+"\n")

    return np.array(IntersecIndex)


def ObtainMultivariateIntersectInSynthetic_AllClusters(dataset, statisticalTests, clusters, IndexAndClass, num_classes, attentionLayer,isTraining):
    # statisticalTest: List of columns with statistical difference. 1.p-value, 2.Cluster step, 3. column of the heatmap        
    stsTest=statisticalTests[0]
    IntersecIndex=[]
    for nClust, clusterToProcess in enumerate(clusters):      
        IntersecIndex.append([])
        if isTraining and (nClust==0 or nClust==2):
            continue           
           
        # Create directory for this set of images.        
        thisFolder = dataset.bioInsights_dir+'/InterpretabilityQuant/BlueIsCluster_GreenIsGroundTruth_ClusterLevel_{}'.format(clusters[nClust])
        if not os.path.exists(thisFolder):
            os.makedirs(thisFolder)

        # For each image.
        for count, subject_info in enumerate(IndexAndClass): 
            statisticalTest =  statisticalTests[count]
            
            # Apply mask to patch
            for patchIDX in range(dataset.findLastIndex(subject_info[1])+1):                       
                
                # Pixel-to-cluster Matrix
                clust0 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], patchIDX, clusters[0]))                 
                if nClust>0:
                    clust1 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], patchIDX, clusters[1]))                                    
                if nClust>1:
                    clust2 = np.load(dataset.processed_dir_cell_types+'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(subject_info[1], clusters[2]))                    
                    clust2 = np.matmul(clust1,clust2)
                clust = clust0 
                clust = clust1 if nClust==1 else clust
                clust = clust2 if nClust==2 else clust                
                
                # Load patch canvas
                Patch_im = np.zeros(np.load(dataset.root+'Raw_Data/Images/'+subject_info[0]+'.npy').shape[:2])

                 # Load Ground-Truth
                if ('SyntheticV2' in dataset.root) or ('SyntheticV_H2' in dataset.root): # fenotipo 4 y 5 de la region 2
                    Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==2
                elif ('SyntheticV1' in dataset.root) or ('SyntheticV_H1' in dataset.root): # fenotipo 6 de la region 3
                    Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==3
                elif ('SyntheticV3' in dataset.root) or ('SyntheticV_H3' in dataset.root):
                    Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==3
                elif ('SyntheticV4' in dataset.root) or ('SyntheticV_H4' in dataset.root):
                    Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')
                    Ground_Truth = np.logical_or(Ground_Truth==2,Ground_Truth==3)

                # Create Patch Label image using the patch size.                         
                division = np.floor(Patch_im.shape[0]/dataset.patch_size)
                lins = np.repeat(list(range(int(division))), dataset.patch_size)
                lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
                for y_indx, y in enumerate(range(int(division))):
                    Patch_im[:int(division*dataset.patch_size),y_indx*dataset.patch_size:(y_indx+1)*dataset.patch_size] = lins+int(y_indx*division)
                Patch_im = Patch_im.astype(int)
                if ('V_H' in dataset.root) or ('V3' in dataset.root) or ('V4' in dataset.root) or ('V2' in dataset.root) or ('V1' in dataset.root):
                    Patch_im = np.transpose(Patch_im)

                #
                cell_type_top1 = np.apply_along_axis(np.argmax, axis=1, arr=clust)                        
                
                # Initialize values
                Thresholds=[0,75,95]
                bestintersection=[-1,-1]
                IntersecIndex[nClust].append(0)
                for cluster_idx in range(clusterToProcess):
                    for threshold in Thresholds:
                        AllClusters = copy.deepcopy(cell_type_top1)*0
                        
                        # Iterate through all significant clusters from this ClusterLevel
                        if 'Synthetic' in dataset.root:                         
                            AllClusters += cluster_idx==cell_type_top1
                            AllClusters[clust[:,cluster_idx]<threshold/100]=0
                        Patch_im=AllClusters[Patch_im]                       
                        
                        # Calculate Intersection Index
                        intersection = np.logical_and(Patch_im, Ground_Truth)                    
                        if intersection.sum():
                            result = intersection.sum() / (1e-16+float(Patch_im.sum()))
                            bestintersection.append(result)
                                                    
                            # In case its the best threshold so far.
                            if result>max(bestintersection[:-1]):
                                IntersecIndex[nClust][count] = max(bestintersection)
                                # Save GT and Image
                                RGBImage = np.zeros((Ground_Truth.shape[0],Ground_Truth.shape[1],3))
                                RGBImage[:,:,1] = Ground_Truth
                                RGBImage[:,:,2] = Patch_im
                                if not isTraining:
                                    plt.imsave(thisFolder+'/IntersectIdx{}_Slide{}_Patch{}.png'.format(result,subject_info[1],patchIDX), RGBImage) 

                        else:
                            bestintersection.append(0)
                            IntersecIndex[nClust][count] = max(bestintersection)
                
        # Save Statistics Information
        if len(IntersecIndex[-1])==0:
            IntersecIndex[-1]=[-1]
        result_int = np.array(IntersecIndex[-1])
        eval_info = {'Intersect-Index': str(sum(result_int[result_int>0])/len(result_int[result_int>0]))+'+-'+str(st.pstdev(result_int[result_int>0])), ' Stats':statisticalTest}

        # Save epoch information into a log file.
        fn = thisFolder+"/Statistics.txt"   
        print(str(fn))
        with open(fn, "w") as myfile:
            myfile.write(str(eval_info)+"\n")

    # Save epoch information into a log file.
    Int_Index = np.array(IntersecIndex).max(0)
    eval_info = {'Intersect-Index': str(np.mean(Int_Index[Int_Index>0]))+'+-'+str(np.std(Int_Index[Int_Index>0])), ' Stats':statisticalTest}
    fn = dataset.bioInsights_dir+'InterpretabilityQuant/Interpretability_Acuracy.txt'                
    print(str(fn))
    with open(fn, "w") as myfile:
        myfile.write(str(eval_info)+"\n")

    return np.array(IntersecIndex)


def InterpretabilityAccuracy_CohortWiseRelevantTME_PerPatch(dataset, MostRelevantTME, clusters, IndexAndClass, num_classes, attentionLayer,isTraining):
    # statisticalTest: List of columns with statistical difference. 1.p-value, 2.Cluster step, 3. column of the heatmap        
    IntersecIndex=[]

    # Interpretability metrics per patch
    Interp_Metrics = {'n_Patches':[], 'Correct_patches':[]}

    # N clusters, and cluster.
    RelTME = MostRelevantTME[4][0][1]
    if 'P' in RelTME:
        nClust = 0
    elif 'N' in RelTME:
        nClust = 1
    elif 'A' in RelTME:
        nClust = 2
    clusterToProcess = int(RelTME[1:])-1
            
    # Create directory for this set of images.        
    thisFolder = dataset.bioInsights_dir+'/InterpretabilityQuant/BlueIsCluster_GreenIsGroundTruth_ClusterLevel_{}'.format(clusters[0])
    if not os.path.exists(thisFolder):
        os.makedirs(thisFolder)

    # For each image.
    for count, subject_info in enumerate(IndexAndClass): 
        
        # Apply mask to patch
        for patchIDX in range(dataset.findLastIndex(subject_info[1])+1):                       
            
            # Pixel-to-cluster Matrix
            clust0 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], patchIDX, clusters[0]))                 
            if nClust>0:
                clust1 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], patchIDX, clusters[1]))                                    
            if nClust>1:
                clust2 = np.load(dataset.processed_dir_cell_types+'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(subject_info[1], clusters[2]))                    
                clust2 = np.matmul(clust1,clust2)
            clust = clust0 
            clust = clust1 if nClust==1 else clust
            clust = clust2 if nClust==2 else clust                
            
            # Load patch canvas
            Patch_im = np.zeros(np.load(dataset.root+'Raw_Data/Images/'+subject_info[0]+'.npy').shape[:2])

                # Load Ground-Truth
            if ('SyntheticV2' in dataset.root) or ('SyntheticV_H2' in dataset.root): # fenotipo 4 y 5 de la region 2
                Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==2
            elif ('SyntheticV1' in dataset.root) or ('SyntheticV_H1' in dataset.root): # fenotipo 6 de la region 3
                Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==3
            elif ('SyntheticV3' in dataset.root) or ('SyntheticV_H3' in dataset.root):
                Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==3
            elif ('SyntheticV4' in dataset.root) or ('SyntheticV_H4' in dataset.root):
                Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')
                Ground_Truth = np.logical_or(Ground_Truth==2,Ground_Truth==3)

            # Create Patch Label image using the patch size.                         
            division = np.floor(Patch_im.shape[0]/dataset.patch_size)
            lins = np.repeat(list(range(int(division))), dataset.patch_size)
            lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
            for y_indx, y in enumerate(range(int(division))):
                Patch_im[:int(division*dataset.patch_size),y_indx*dataset.patch_size:(y_indx+1)*dataset.patch_size] = lins+int(y_indx*division)
            Patch_im = Patch_im.astype(int)
            if ('V_H' in dataset.root) or ('V3' in dataset.root) or ('V4' in dataset.root) or ('V2' in dataset.root) or ('V1' in dataset.root):
                Patch_im = np.transpose(Patch_im)

            #
            cell_type_top1 = np.apply_along_axis(np.argmax, axis=1, arr=clust)                        
            
            # Initialize values
            Thresholds=[0,25,75,95,96,97,98,99]
            bestintersection=[-1,-1]    
            Interp_Metrics['n_Patches'].append(0)        
            Interp_Metrics['Correct_patches'].append(0)        
            # for cluster_idx in range(clusterToProcess):
            for threshold in Thresholds:
                AllClusters = copy.deepcopy(cell_type_top1)*0
                Patch_im_Int = copy.deepcopy(Patch_im)

                # Iterate through all significant clusters from this ClusterLevel
                if 'Synthetic' in dataset.root:                         
                    AllClusters += clusterToProcess==cell_type_top1
                    AllClusters[clust[:,clusterToProcess]<threshold/100]=0
                Patch_im_Int=AllClusters[Patch_im_Int]                       
                
                # Calculate Intersection Index
                intersection = np.logical_and(Patch_im_Int, Ground_Truth)                    
                if intersection.sum():
                    result = intersection.sum() / (1e-16+float(Patch_im_Int.sum()))
                    bestintersection.append(result)
                                            
                    # In case its the best threshold so far.
                    if result>max(bestintersection[:-1]):
                        # Extract interpretability
                        Interp_Metrics['n_Patches'][-1]=Patch_im_Int.sum()
                        Interp_Metrics['Correct_patches'][-1]=intersection.sum()
                        # Save GT and Image
                        RGBImage = np.zeros((Ground_Truth.shape[0],Ground_Truth.shape[1],3))
                        RGBImage[:,:,1] = Ground_Truth
                        RGBImage[:,:,2] = Patch_im_Int
                        if not isTraining:
                            plt.imsave(thisFolder+'/IntersectIdx{}_Slide{}_Patch{}.png'.format(result,subject_info[1],patchIDX), RGBImage) 

                else:
                    bestintersection.append(0)
            
    # Save Statistics Information    
    print(['Num Patches: ',str(sum(Interp_Metrics['n_Patches'])+1)])
    result = sum(Interp_Metrics['Correct_patches'])/(sum(Interp_Metrics['n_Patches'])+1)
    eval_info = {'Intersect-Index': str(result), ' Stats':MostRelevantTME}

    # Save epoch information into a log file.
    fn = thisFolder+"/Statistics.txt"   
    print(str(fn))
    with open(fn, "w") as myfile:
        myfile.write(str(eval_info)+"\n")

    return result


def InterpretabilityAccuracy_ImageWiseRelevantTME_PerPatch(dataset, MostRelevantTME, clusters, IndexAndClass, num_classes, attentionLayer,isTraining):
    # statisticalTest: List of columns with statistical difference. 1.p-value, 2.Cluster step, 3. column of the heatmap        
    IntersecIndex=[]

    # Interpretability metrics per patch
    Interp_Metrics = {'n_Patches':[], 'Correct_patches':[]}
            
    # For each image.
    for count, subject_info in enumerate(IndexAndClass): 
        
        nClust = clusters.index(int(MostRelevantTME[count][2][0]))
        clusterToProcess = MostRelevantTME[count][2][1]
        # Create directory for this set of images.        
        thisFolder = dataset.bioInsights_dir+'/InterpretabilityQuant/BlueIsCluster_GreenIsGroundTruth_ClusterLevel_{}'.format(clusters[nClust])
        if not os.path.exists(thisFolder):
            os.makedirs(thisFolder)

        # Apply mask to patch
        for patchIDX in range(dataset.findLastIndex(subject_info[1])+1):                       
            
            # Pixel-to-cluster Matrix
            clust0 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], patchIDX, clusters[0]))                 
            if nClust>0:
                clust1 = np.load(dataset.processed_dir_cell_types+'cluster_assignmentPerPatch_Index_{}_{}_ClustLvl_{}.npy'.format(subject_info[1], patchIDX, clusters[1]))                                    
            if nClust>1:
                clust2 = np.load(dataset.processed_dir_cell_types+'cluster_assignment_Index_{}_ClustLvl_{}.npy'.format(subject_info[1], clusters[2]))                    
                clust2 = np.matmul(clust1,clust2)
            clust = clust0 
            clust = clust1 if nClust==1 else clust
            clust = clust2 if nClust==2 else clust                
            
            # Load patch canvas
            Patch_im = np.zeros(np.load(dataset.root+'Raw_Data/Images/'+subject_info[0]+'.npy').shape[:2])

                # Load Ground-Truth
            if ('SyntheticV2' in dataset.root) or ('SyntheticV_H2' in dataset.root): # fenotipo 4 y 5 de la region 2
                Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==2
            elif ('SyntheticV1' in dataset.root) or ('SyntheticV_H1' in dataset.root): # fenotipo 6 de la region 3
                Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==3
            elif ('SyntheticV3' in dataset.root) or ('SyntheticV_H3' in dataset.root):
                Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')==3
            elif ('SyntheticV4' in dataset.root) or ('SyntheticV_H4' in dataset.root):
                Ground_Truth = np.load(dataset.root+'Raw_Data/Experiment_Information/Synthetic_Ground_Truth/Ground_Truth_'+subject_info[0][6:]+'.npy')
                Ground_Truth = np.logical_or(Ground_Truth==2,Ground_Truth==3)

            # Create Patch Label image using the patch size.                         
            division = np.floor(Patch_im.shape[0]/dataset.patch_size)
            lins = np.repeat(list(range(int(division))), dataset.patch_size)
            lins = np.repeat(np.expand_dims(lins,axis=1), dataset.patch_size, axis=1)
            for y_indx, y in enumerate(range(int(division))):
                Patch_im[:int(division*dataset.patch_size),y_indx*dataset.patch_size:(y_indx+1)*dataset.patch_size] = lins+int(y_indx*division)
            Patch_im = Patch_im.astype(int)
            if ('V_H' in dataset.root) or ('V3' in dataset.root) or ('V4' in dataset.root) or ('V2' in dataset.root) or ('V1' in dataset.root):
                Patch_im = np.transpose(Patch_im)

            #
            cell_type_top1 = np.apply_along_axis(np.argmax, axis=1, arr=clust)                        
            
            # Initialize values
            Thresholds=[95,96,97,98,99]
            bestintersection=[-1,-1]    
            Interp_Metrics['n_Patches'].append(0)        
            Interp_Metrics['Correct_patches'].append(0)        
            # for cluster_idx in range(clusterToProcess):
            for threshold in Thresholds:
                AllClusters = copy.deepcopy(cell_type_top1)*0
                Patch_im_Int = copy.deepcopy(Patch_im)

                # Iterate through all significant clusters from this ClusterLevel
                if 'Synthetic' in dataset.root:                         
                    AllClusters += clusterToProcess==cell_type_top1
                    AllClusters[clust[:,clusterToProcess]<threshold/100]=0
                Patch_im_Int=AllClusters[Patch_im_Int]                       
                
                # Calculate Intersection Index
                intersection = np.logical_and(Patch_im_Int, Ground_Truth)                    
                if intersection.sum():
                    result = intersection.sum() / (1e-16+float(Patch_im_Int.sum()))
                    bestintersection.append(result)
                                            
                    # In case its the best threshold so far.
                    if result>max(bestintersection[:-1]):
                        # Extract interpretability
                        Interp_Metrics['n_Patches'][-1]=Patch_im_Int.sum()
                        Interp_Metrics['Correct_patches'][-1]=intersection.sum()
                        # Save GT and Image
                        RGBImage = np.zeros((Ground_Truth.shape[0],Ground_Truth.shape[1],3))
                        RGBImage[:,:,1] = Ground_Truth
                        RGBImage[:,:,2] = Patch_im_Int
                        if not isTraining:
                            plt.imsave(thisFolder+'/IntersectIdx{}_Slide{}_Patch{}.png'.format(result,subject_info[1],patchIDX), RGBImage) 

                else:
                    bestintersection.append(0)
        
    # Save Statistics Information    
    result = sum(Interp_Metrics['Correct_patches'])/(sum(Interp_Metrics['n_Patches'])+1)
    eval_info = {'Intersect-Index': str(result), ' Stats':MostRelevantTME}

    # Save epoch information into a log file.
    fn = thisFolder+"/Statistics.txt"   
    print(str(fn))
    with open(fn, "w") as myfile:
        myfile.write(str(eval_info)+"\n")

    return result


import torch
import NaroNet.utils.utilz
import numpy as np
EPS = 1e-15

def Sparse_Pooling(y, x, adj, s,device,args,model):
    
    x = x.unsqueeze(0) if x.dim() == 2 else x
    adj = adj.unsqueeze(0) if adj.dim() == 2 else adj
    s = s.unsqueeze(0) if s.dim() == 2 else s

    batch_size, num_nodes, _ = x.size()

    out = (torch.matmul(s.transpose(1, 2), x).transpose(1,2).transpose(0,1)/(s.sum(-2)+1e-16)).transpose(0,1).transpose(1,2)  # /(num_nodes/s.shape[-1])
    out_adj = torch.matmul(torch.matmul(s.transpose(1, 2), adj), s)
    # out_adj = out_adj/(out_adj.sum(dim=-1,keepdim=True)+1e-16)        
    # out_adj = out_adj/(out_adj.sum(dim=-2,keepdim=True)+1e-16)

    
    D = out_adj.sum(dim=-1,keepdim=True).pow(-0.5)
    D[D == float('inf')] = 0
    out_adj = out_adj*D
    out_adj = torch.transpose(torch.transpose(out_adj,1,2)*D,1,2)

    minCUT_loss = 0 

    return out, out_adj, minCUT_loss

def Dense_Pooling(x, edge_index, s, device, num_nodes,obtainOutX):
        
    
    # Calculate Coarsened Graph
    out_x = torch.zeros(s.shape[0],s.shape[2],x.shape[2],dtype=torch.float32).to(device)
    out_adj = torch.zeros(s.shape[0],s.shape[2],s.shape[2],dtype=torch.float32).to(device)
    if obtainOutX:
        for i in range(s.shape[0]): 
            # out_x[i,:,:] = torch.matmul(s[i,:num_nodes[i],:].t(),x[i,:num_nodes[i],:])/(num_nodes[i]/s.shape[-1]) if not args['ObjectiveCluster'] else out_x[i,:,:]
            out_x[i,:,:] = (torch.matmul(s[i,:num_nodes[i],:].t(),x[i,:num_nodes[i],:]).t()/(s[i,:num_nodes[i],:].sum(-2)+1e-16)).t()
            out_adj[i,:,:] = torch.matmul(s[i,:num_nodes[i],:].t(),torch.sparse.mm(edge_index[i].float(),s[i,:,:])[:num_nodes[i],:])+1e-16
        
    # Normalize Graph
    if obtainOutX:
        D = out_adj.sum(dim=-1,keepdim=True).pow(-0.5)
        D[D == float('inf')] = 0
        out_adj = out_adj*D
        out_adj = torch.transpose(torch.transpose(out_adj,1,2)*D,1,2)

    return out_x, out_adj


import argparse
import torch
import torch.backends.cudnn as cudnn
from torchvision import models
from data_aug.contrastive_learning_dataset import ContrastiveLearningDataset,  ContrastiveLearningDataset_PCL
from models.resnet_simclr import ResNetSimCLR
from simclr import SimCLR

model_names = sorted(name for name in models.__dict__
                     if name.islower() and not name.startswith("__")
                     and callable(models.__dict__[name]))

parser = argparse.ArgumentParser(description='PyTorch SimCLR')
parser.add_argument('-data', metavar='DIR', default='./datasets', # '/mnt/g/Mi unidad/Proyectos/NaroNet/Datasets/Endometrial_POLE_toy/'
                    help='path to dataset')
parser.add_argument('-dataset-name', default='stl10',
                    help='dataset name', choices=['stl10', 'cifar10'])
parser.add_argument('-a', '--arch', metavar='ARCH', default='resnet18',
                    choices=model_names,
                    help='model architecture: ' +
                         ' | '.join(model_names) +
                         ' (default: resnet50)')
parser.add_argument('-j', '--workers', default=6, type=int, metavar='N',
                    help='number of data loading workers (default: 32)')
parser.add_argument('--epochs', default=200, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('-b', '--batch-size', default=256, type=int,
                    metavar='N',
                    help='mini-batch size (default: 256), this is the total '
                         'batch size of all GPUs on the current node when '
                         'using Data Parallel or Distributed Data Parallel')
parser.add_argument('--lr', '--learning-rate', default=0.0003, type=float,
                    metavar='LR', help='initial learning rate', dest='lr')
parser.add_argument('--wd', '--weight-decay', default=1e-4, type=float,
                    metavar='W', help='weight decay (default: 1e-4)',
                    dest='weight_decay')
parser.add_argument('--seed', default=None, type=int,
                    help='seed for initializing training. ')
parser.add_argument('--disable-cuda', action='store_true',
                    help='Disable CUDA')
parser.add_argument('--fp16-precision', action='store_true',
                    help='Whether or not to use 16-bit precision GPU training.')

parser.add_argument('--out_dim', default=128, type=int,
                    help='feature dimension (default: 128)')
parser.add_argument('--log-every-n-steps', default=100, type=int,
                    help='Log every n steps')
parser.add_argument('--temperature', default=0.07, type=float,
                    help='softmax temperature (default: 0.07)')
parser.add_argument('--n-views', default=2, type=int, metavar='N',
                    help='Number of views for contrastive learning training.')
parser.add_argument('--gpu-index', default=0, type=int, help='Gpu index.')


def main():
    args = parser.parse_args()
    assert args.n_views == 2, "Only two view training is supported. Please use --n-views 2."
    # check if gpu training is available
    if not args.disable_cuda and torch.cuda.is_available():
        args.device = torch.device('cuda:0')
        cudnn.deterministic = True
        cudnn.benchmark = True
    else:
        args.device = torch.device('cpu')
        args.gpu_index = -1

    dataset = ContrastiveLearningDataset(args.data)
    dataset_PCL = ContrastiveLearningDataset_PCL('/mnt/g/Mi unidad/Proyectos/NaroNet/Datasets/Endometrial_POLE_toy/')

    train_dataset = dataset.get_dataset(args.dataset_name, args.n_views)
    
    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=args.batch_size, shuffle=True,
        num_workers=args.workers, pin_memory=True, drop_last=True)

    model = ResNetSimCLR(base_model=args.arch, out_dim=args.out_dim)

    optimizer = torch.optim.Adam(model.parameters(), args.lr, weight_decay=args.weight_decay)

    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=len(train_loader), eta_min=0,
                                                           last_epoch=-1)

    #  It’s a no-op if the 'gpu_index' argument is a negative integer or None.
    with torch.cuda.device(args.gpu_index):
        simclr = SimCLR(model=model, optimizer=optimizer, scheduler=scheduler, args=args)
        simclr.train(train_loader)


if __name__ == "__main__":
    main()



import torch
import numpy as np
import torch
import torch.nn.functional as F
import gc

def calc_contrast_loss_prev(features,assignment_orig,device,args):
    '''
    Maximize contrast of top1 patches and their neighbors.
    
    '''
    # Use a batch size of 1000
    # TOP 1%
    # K_top = max(min(int(assignment_orig.shape[0]*0.01/assignment_orig.shape[1]),assignment_orig.shape[0]),assignment_orig.shape[1])
    # n_batch = min(100,K_top)
    # n_neighb = min(50,assignment_orig.shape[0])
    # TOP 1
    K_top = 1
    n_batch = max(min(int(assignment_orig.shape[0]*args['Perc_high_conf_patches']*2),assignment_orig.shape[0]),10,assignment_orig.shape[1])
    neg_batch = min(10000,assignment_orig.shape[0])

    # Calculate feature similarities
    features = F.normalize(features.squeeze(), dim=1)    
    assignment_orig = F.normalize(assignment_orig,dim=1)
    temp = 0.07

    loss = 0
    for ph in range(assignment_orig.shape[1]):
        
        # Positives
        sel_vals = assignment_orig[:,ph].topk(K_top,0)[1]   
        feats = torch.matmul(features[sel_vals], features.T)        
        top_elem = (feats.topk(n_batch,1)[1]).squeeze()
        positives = torch.exp((torch.matmul(assignment_orig[top_elem],assignment_orig[top_elem].T)/temp))        

        # Negatives
        sel_vals_neg = torch.randint(0, assignment_orig.shape[0]-1, (neg_batch,))
        negatives = torch.exp((torch.matmul(assignment_orig[top_elem],assignment_orig[sel_vals_neg].T)/temp))        
        loss += -torch.log(positives).mean()*temp
        loss += torch.log(negatives).mean()*temp
        
    return loss/assignment_orig.shape[1]


def calc_contrast_loss_topBottomPerc(features,assignment_orig,device,args):


    # Number of patches that define the phenotype.
    K_top = 1 
    
    # Number of patches going to 
    pos_batch = max(min(int(assignment_orig.shape[0]*args['Perc_high_conf_patches']),assignment_orig.shape[0]),4,assignment_orig.shape[1])

    # Calculate feature similarities
    features = F.normalize(features.squeeze(), dim=1)    
    assignment_orig = F.normalize(assignment_orig,dim=1)
    temp = 0.07

    loss = 0
    for ph in range(assignment_orig.shape[1]):
        
        # Perfect assignment
        objective = torch.zeros(assignment_orig.shape[1],device=device)
        objective[ph] = 1

        # Obtain top-k assignment that defines the TME
        top_assignment = assignment_orig[:,ph].topk(K_top,0)[1]   
        
        # Feature distance to top-k
        feat_dist = torch.matmul(features[top_assignment], features.T)        
        
        # Select most similar patches
        top_elem = (feat_dist.topk(pos_batch,1)[1]).squeeze()
        
        # Positive loss - patches with similar features to the top-k patches should have the same assignment
        positives = torch.exp((torch.matmul(objective,assignment_orig[top_elem].T)/temp)) 
        loss += -torch.log(positives).mean()*temp       

        # combine feature dist and assignment
        # highest_conf_assignment = (((1-feat_dist)/2)**2 + (assignment_orig[:,ph])**2).topk(pos_batch,1)[1]   
        # negatives = torch.exp((torch.matmul(objective,assignment_orig[highest_conf_assignment.squeeze()].T)/temp))    
        # loss += (torch.log(negatives)).mean()*temp
        

        # Patches with highest confidence assignment 
        highest_conf_assignment = assignment_orig[:,ph].topk(pos_batch*2,0)[1]           
        # Patch with highest features distance from top-1
        bottom_val = feat_dist[:,highest_conf_assignment].topk(pos_batch,1,largest=False)[1].squeeze()
        # Negative loss - patches with dissimilar features to the top-k patches but with high assignment confidence should have a dissimilar assignment.
        negatives = torch.exp((torch.matmul(objective,assignment_orig[highest_conf_assignment[bottom_val]].T)/temp))        
        loss += (torch.log(negatives)*(1-feat_dist[0,highest_conf_assignment[bottom_val]])/2).mean()*temp
        
    return loss/assignment_orig.shape[1]

def calc_contrast_loss(features,assignment_orig,device,args):



    # Number of patches going to 
    pos_batch = max(int(assignment_orig.shape[0]*args['Perc_high_conf_patches']),1)

    # Calculate feature similarities
    features = F.normalize(features.squeeze(), dim=1)    
    assignment_orig = F.normalize(assignment_orig,dim=1)
    temp = 0.07

    loss = 0
    for ph in range(assignment_orig.shape[1]):
        
        # Perfect assignment
        objective = torch.zeros(assignment_orig.shape[1],device=device)
        objective[ph] = 1

        # Obtain top-k assignment that defines the TME
        top_assignment = assignment_orig[:,ph].topk(pos_batch*10,0)[1]   
        
        # Feature distance to top-k
        feat_dist = torch.matmul(features[top_assignment], features[top_assignment].T)
        
        # Select most similar patches
        prototype = (feat_dist.sum(0).topk(1,0)[1])
        top_elem = (feat_dist[prototype][0,:].topk(pos_batch,0)[1])
        
        # Positive loss - patches with similar features to the top-k patches should have the same assignment
        positives = torch.exp((torch.matmul(objective,assignment_orig[top_assignment][top_elem].T)/temp)) 

        # loss += -torch.log(positives).mean()*temp       
        

        # combine feature dist and assignment
        # highest_conf_assignment = (((1-feat_dist)/2)**2 + (assignment_orig[:,ph])**2).topk(pos_batch,1)[1]   
        # negatives = torch.exp((torch.matmul(objective,assignment_orig[highest_conf_assignment.squeeze()].T)/temp))    
        # loss += (torch.log(negatives)).mean()*temp
        

        # Patches with highest confidence assignment          
        bottom_val = feat_dist[prototype][0,:].topk(pos_batch*9,0,largest=False)[1]
        sel_vals_neg = torch.randint(0, pos_batch*9, (pos_batch,))
        # Negative loss - patches with dissimilar features to the top-k patches but with high assignment confidence should have a dissimilar assignment.
        # negatives = torch.exp((torch.matmul(objective,assignment_orig[top_assignment][bottom_val[sel_vals_neg]].T)/temp))    
        # loss += torch.log(negatives).mean()*temp    
        loss += -torch.log(positives).mean()*temp
        # print('Positives: '+str((torch.log(positives).mean()*temp).item()*100)[:5]+' feat_pos: '+str(feat_dist[prototype][0,top_elem[-1]].item())[:6]+'   -   Negative: '+str((torch.log(negatives).mean()*temp).item()*100)[:5]+' feat_neg: '+str(feat_dist[prototype][0,bottom_val[0]].item())[:6])
        
    return loss/assignment_orig.shape[1]


def calc_patch_ent(s,device):    
    return ((-s/s.sum(dim=-1,keepdim=True) * torch.log(s/s.sum(dim=-1,keepdim=True) + 1e-15)).sum(dim=-1)/torch.log(torch.tensor(s.shape[1],dtype=torch.float32,device=device))).mean()   

def patch_contrast_entropy(x, s, num_nodes, device,args):
    
    # Calculate contrast acc        
    contrast_loss = calc_contrast_loss(torch.cat([x[i,:num_nodes[i],:] for i in range(s.shape[0])]), torch.cat([s[i,:num_nodes[i],:] for i in range(s.shape[0])]),device,args)          
    
    # Calculate Patch Entropy
    PatchEntropy_loss = calc_patch_ent(torch.cat([s[i,:num_nodes[i],:] for i in range(s.shape[0])]),device)                     

    return contrast_loss, PatchEntropy_loss
    
def Patient_entropy(args,S,device):
    # Keep High Entropy in the network
    if args['Max_Pat_Entropy']:                                          
        NN_loss1 = ((S[0]/S[0].sum(-1,keepdim=True)*torch.log(S[0]/S[0].sum(-1,keepdim=True)+ 1e-15)).sum(dim=1)/torch.log(torch.tensor(S[0].shape[1],dtype=torch.float32,device=device))).mean()
        NN_loss2 = ((S[1]/S[1].sum(-1,keepdim=True)*torch.log(S[1]/S[1].sum(-1,keepdim=True)+ 1e-15)).sum(dim=1)/torch.log(torch.tensor(S[1].shape[1],dtype=torch.float32,device=device))).mean()
        NN_loss3 = ((S[2]/S[2].sum(-1,keepdim=True)*torch.log(S[2]/S[2].sum(-1,keepdim=True)+ 1e-15)).sum(dim=1)/torch.log(torch.tensor(S[2].shape[1],dtype=torch.float32,device=device))).mean()
        print('Entropy Phenotypes: ',NN_loss1.item(),' Entropy TissueCommunities: ',NN_loss2.item(), 'Entropy TissueCommunitiesInter: ',NN_loss3.item(),'Min_Clust: ',S[0].sum(-2).min().item())
        return NN_loss1, NN_loss2, NN_loss3
    else:
        return 0, 0, 0

def NearestNeighborClassification(self):
    NN_loss, _, _, _ = utilz.nearestNeighbor_loss(self.training, args, self.sPatient[-1], data.y.long(), self.NNCPosition,index,Indices,labels,device,self)                                
    NN_loss = -(-torch.cat((self.S[0],self.S[1],self.S[2]),dim=1) * torch.log(torch.cat((self.S[0],self.S[1],self.S[2]),dim=1) + 1e-15)).sum(dim=-1).mean()                

def f_test_loss(features,labels,device,lin1,lin2,BN,args):
    
    if args['F-test']:
        # Unrestricted Model
        UM = F.relu(lin1(features)) # torch.cat(self.X,dim=-1)            
        UM = lin2(UM)         
        # Model Loss of Unrestricted Model
        unrestricted_loss = F.cross_entropy(UM, labels[:,0].long())     
        pred = UM.max(1)[1].detach().cpu().numpy()
        print('train_acc:',np.equal(pred,labels[:,0].cpu().numpy()).mean())

        # # Sum of Squares Residuals of unrestricted model
        # unrestricted_RSS = torch.pow((torch.matmul(features,regression_model.weight.t())+regression_model.bias).squeeze()-labels,2).sum()    
        # # Leave-one-out f-test model
        restricted_loss = torch.zeros(features.shape[1],device=device)
        
        # For copies
        # W_f = regression_model.weight.clone()
        indices = torch.arange(features.shape[1],device=device)
        
        # Restrict model per feature.
        for f in range(features.shape[1]):
            # Put Weight of f_index to zero        
            # regression_model.weight[0,f] = 0

            # Obtain Sum of Squares of restricted model        
            Restricted_Info = torch.zeros(features.shape,device=device)        
            Restricted_Info[:,indices!=f] = features[:,indices!=f]
            restricted_loss[f] = F.cross_entropy(lin2(BN(F.relu(lin1(Restricted_Info)))), labels[:,0].long()) if Restricted_Info.shape[0]>1 else F.cross_entropy(lin2(F.relu(lin1(Restricted_Info))), labels[:,0].long())        
            # torch.cat(self.X,dim=-1)                 
            
            
            # restricted_RSS = torch.pow((torch.matmul(features[:,x[x!=f]],regression_model.weight[:,x[x!=f]].t())+regression_model.bias).squeeze()-labels,2).sum()

            # F-test
            # loss[f] = (restricted_RSS-unrestricted_RSS)/(features.shape[0]-features.shape[1])

        # loss = F.relu(loss)
        # Sparsify the unrestricted models
        restricted_loss = (restricted_loss/unrestricted_loss)  
        restricted_loss = restricted_loss/restricted_loss.sum()  
        restricted_loss = -(restricted_loss * torch.log(restricted_loss + 1e-15)).mean()
        # restricted_loss = -restricted_loss.max()
        # restricted_loss = -(torch.sqrt(restricted_loss.shape[0]*torch.ones(1,device=device))-(restricted_loss.sum()/(torch.pow(restricted_loss,2).sum()+1e-16)))/(torch.sqrt(restricted_loss.shape[0]*torch.ones(1,device=device))-1)
        print('sparsify the f-test',restricted_loss, 'loss of original Model',unrestricted_loss)
        f_test_loss = unrestricted_loss+restricted_loss
    else:
        f_test_loss = 0

    return f_test_loss

def SupConLoss(features,labels,device,temperature=0.07):
    """Supervised Contrastive Learning: https://arxiv.org/pdf/2004.11362.pdf.
    it degenerates to SimCLR unsupervised loss:
    https://arxiv.org/pdf/2002.05709.pdf
    Args:
        features: hidden vector of shape [bsz, n_views, ...].
        labels: ground truth of shape [bsz].
        mask: contrastive mask of shape [bsz, bsz], mask_{i,j}=1 if sample j
            has the same class as sample i. Can be asymmetric.
    Returns:
        A loss scalar.
    """
    # device = (torch.device('cuda')
    #           if features.is_cuda
    #           else torch.device('cpu'))
    
    loss = torch.zeros(1,device=device)
    features = features.unsqueeze(dim=1)
    Contrast_acc = 0
    n_iter = 0

    # Calculate loss only if there are repeated values in loss
    a, rep = np.unique(labels.cpu().numpy(),return_counts=True)
    if any(rep>1):
        # Iterate Over each patient i.        
        for i in range(features.shape[0]): 
            
            # Reinitialize patient_i_loss
            patient_i_loss = torch.zeros(1,device=device)

            # Iterate Over each patient j.
            for j in range(features.shape[0]): 
                
                # Select patient if they have same labels.
                if (i!=j) and (labels[i]==labels[j]):                 
                    # Apply supContrast Loss
                    patient_i_loss += torch.log(
                        torch.exp(F.cosine_similarity(features[i],features[j])/temperature)/
                        torch.exp(F.cosine_similarity(features[i],features[labels[i]!=labels].squeeze(dim=1))/temperature).mean())
                    Contrast_acc += F.cosine_similarity(features[i],features[j])-F.cosine_similarity(features[i],features[labels[i]!=labels].squeeze(dim=1)).mean()
                    n_iter += 1 

            # Normalize patient_i_loss
            loss += (-torch.ones(1,device=device)/((labels[i]==labels).sum()-1+1e-16))*patient_i_loss
        
        loss /= features.shape[0]
                    
    return loss, Contrast_acc/(n_iter+1e-15)

def SupConLoss_Total(args,S,labels,data,device):
    if args['orthoColor']:
        # supconloss = torch.relu(self.SUPCONlin(torch.cat((self.S[0],self.S[1],self.S[2]),dim=1)))
        # supconloss = utilz.SupConLoss(torch.cat((self.S[0],self.S[1],self.S[2]),dim=1),data.y,device,temperature=0.07)
        if args['F-test']:
            supconloss = SupConLoss(S[0][:,:max(labels)+1],data.y,device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[0][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
            supconloss += SupConLoss(S[1][:,:max(labels)+1],data.y,device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[1][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
            supconloss += SupConLoss(S[2][:,:max(labels)+1],data.y,device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[2][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
            # f_test_loss = -self.S[0][:,max(labels)+1:].mean()
            # f_test_loss += -self.S[1][:,max(labels)+1:].mean()
            # f_test_loss += -self.S[2][:,max(labels)+1:].mean()
            # f_test_loss = f_test_loss*0.001
        else:
            supconloss1, sup_acc1  = SupConLoss(S[0],data.y[:,0],device,temperature=0.07)
            supconloss2, sup_acc2 = SupConLoss(S[1],data.y[:,0],device,temperature=0.07)
            supconloss3, sup_acc3 = SupConLoss(S[2],data.y[:,0],device,temperature=0.07)
            if np.argmin([sup_acc1,sup_acc2,sup_acc3])==0:
                supconloss=supconloss1
            elif np.argmin([sup_acc1,sup_acc2,sup_acc3])==1:
                supconloss=supconloss2
            else:
                supconloss=supconloss3
            print('Supervised_Contrast_accuracy: ',(sup_acc1).item(),+sup_acc2.item(), sup_acc3.item())
    else:
        supconloss=0

    return supconloss

def unsup_ColorLoss(projection,labels, features,device,temperature=0.07):
    """Supervised Contrastive Learning: https://arxiv.org/pdf/2004.11362.pdf.
    it degenerates to SimCLR unsupervised loss:
    https://arxiv.org/pdf/2002.05709.pdf
    Args:
        features: hidden vector of shape [bsz, n_views, ...].
        labels: ground truth of shape [bsz].
        mask: contrastive mask of shape [bsz, bsz], mask_{i,j}=1 if sample j
            has the same class as sample i. Can be asymmetric.
    Returns:
        A loss scalar.
    """
    # device = (torch.device('cuda')
    #           if features.is_cuda
    #           else torch.device('cpu'))
    
    loss = torch.zeros(1,device=device)
    projection = projection/projection.max()
    projection = projection.unsqueeze(dim=1)
    features = features/features.max()
    features = features.unsqueeze(dim=1)

    unsup_Contrast_acc = 0 
    n_iter = 0

    # Calculate loss only if there are more than One Image.
    if labels.shape[0]>2:
        # Iterate Over each patient i.        
        for i in list(range(0,projection.shape[0],1)): 
            
            # Reinitialize patient_i_loss
            patient_i_loss = torch.zeros(1,device=device)

            # Iterate Over each patient j.
            for j in range(projection.shape[0]): 
                
                # Select patient if they have same labels.
                if (i!=j):                 
                    # Apply supContrast Loss
                    patient_i_loss += torch.exp(F.cosine_similarity(projection[i],projection[j]))
                    # unsup_Contrast_acc += F.cosine_similarity(features[i],features[i+1])-F.cosine_similarity(features[i],features[j])
                    n_iter += 1 

            # Normalize patient_i_loss
            loss += patient_i_loss
        
        loss /= projection.shape[0]*(projection.shape[0]-1)
                    
    return loss, unsup_Contrast_acc/(n_iter+1e-16)

def UnsupConLoss(projection,labels, features,device,temperature=0.07):
    """Supervised Contrastive Learning: https://arxiv.org/pdf/2004.11362.pdf.
    it degenerates to SimCLR unsupervised loss:
    https://arxiv.org/pdf/2002.05709.pdf
    Args:
        features: hidden vector of shape [bsz, n_views, ...].
        labels: ground truth of shape [bsz].
        mask: contrastive mask of shape [bsz, bsz], mask_{i,j}=1 if sample j
            has the same class as sample i. Can be asymmetric.
    Returns:
        A loss scalar.
    """
    # device = (torch.device('cuda')
    #           if features.is_cuda
    #           else torch.device('cpu'))
    
    loss = torch.zeros(1,device=device)

    projection = projection.unsqueeze(dim=1)
    features = features.unsqueeze(dim=1)

    unsup_Contrast_acc = 0 
    n_iter = 0

    # Calculate loss only if there are more than One Image.
    if labels.shape[0]>2:
        # Iterate Over each patient i.        
        for i in list(range(0,projection.shape[0],2)): 
            
            # Reinitialize patient_i_loss
            patient_i_loss = torch.zeros(1,device=device)

            # Iterate Over each patient j.
            for j in range(projection.shape[0]): 
                
                # Select patient if they have same labels.
                if (i!=j) and ((i+1)!=j):                 
                    # Apply supContrast Loss
                    patient_i_loss += -torch.log(
                        torch.exp(F.cosine_similarity(projection[i],projection[i+1])/temperature)/
                        torch.exp(F.cosine_similarity(projection[i],projection[j])/temperature).mean())
                    unsup_Contrast_acc += F.cosine_similarity(features[i],features[i+1])-F.cosine_similarity(features[i],features[j])
                    n_iter += 1 

            # Normalize patient_i_loss
            loss += patient_i_loss
        
        loss /= projection.shape[0]*(projection.shape[0]-2)
                    
    return loss, unsup_Contrast_acc/(n_iter+1e-16)

def UnsupConLoss_Total(args, S, data, device, self):
    if args['UnsupContrast'] and self.training:
        # unsupconloss1, unsup_acc1 = utilz.UnsupConLoss(self.lin1_unsupB(F.relu(self.lin1_unsupA(self.S[0]))),data.y,self.S[0],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[0][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        # unsupconloss2, unsup_acc2 = utilz.UnsupConLoss(self.lin2_unsupB(F.relu(self.lin2_unsupA(self.S[1]))),data.y, self.S[1],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[1][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        # unsupconloss3, unsup_acc3 = utilz.UnsupConLoss(self.lin3_unsupB(F.relu(self.lin3_unsupA(self.S[2]))),data.y,self.S[2],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[2][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        unsupconloss1, unsup_acc1 = UnsupConLoss(S[0],data.y,S[0],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[0][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        unsupconloss2, unsup_acc2 = UnsupConLoss(S[1],data.y,S[1],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[1][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                
        unsupconloss3, unsup_acc3 = UnsupConLoss(S[2],data.y,S[2],device,temperature=0.07)# + torch.stack([-(i * torch.log(i + 1e-15)).sum() for i in self.S[2][:,:max(labels)+1]/self.S[0][:,:max(labels)+1].sum(-1).unsqueeze(-1)]).mean()                    
        # if np.argmin([unsup_acc1,unsup_acc2,unsup_acc3])==0:
        #     unsupconloss=unsupconloss1
        # elif np.argmin([unsup_acc1,unsup_acc2,unsup_acc3])==1:
        #     unsupconloss=unsupconloss2
        # else:
        #     unsupconloss=unsupconloss2
        unsup_acc = (unsup_acc1+unsup_acc2+unsup_acc3)/3
        # else:
        #     unsupconloss=unsupconloss3
        # print('Unsupervised_accuracy: ',(unsup_acc1).item(),+unsup_acc2.item(), unsup_acc3.item())
        print('Unsupervised_accuracy: ',unsup_acc)
        return unsupconloss1, unsup_acc1, unsupconloss2, unsup_acc2, unsupconloss3, unsup_acc3 
    else: 
        return 0,0,0,0,0,0

def Lasso_Feat_Selection(layer1,layer2):
    return 0.1*layer1.weight.abs().sum() + 0.1*layer2.weight.abs().sum()


    




import logging
import os
import sys

import torch
import torch.nn.functional as F
from torch.cuda.amp import GradScaler, autocast
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from NaroNet.Patch_Contrastive_Learning.utils import save_config_file, accuracy, save_checkpoint
import pandas as pd
import numpy as np
from PIL import Image
from sklearn.cluster import AgglomerativeClustering
import gc 
from openTSNE import TSNE
from sklearn.preprocessing import StandardScaler
from matplotlib import pyplot as plt
import seaborn as sns
import sklearn
gc.set_threshold(0,0,0)

torch.manual_seed(0)

def visualize_using_clustering_ALL(dataset, out_dir, features, features_idx, patch_images):

    # Cluster image using all information.
    clusters_ALL = sklearn.cluster.KMeans(n_clusters=5).fit(features[:,2:])
    
    # Calculate image size
    for im_idx, name in enumerate(patch_images):
        
        # # CLusters all
        sel_feats = features[features_idx==im_idx]
        image_size = sel_feats[:,:2].max(0)+sel_feats[:,:2].min(0)
        patch_size = int((sel_feats[:,:2].min(0)*2)[0])
        sel_clusters_all = clusters_ALL.labels_[features_idx==im_idx]

        # CLustering of this image
        clusters_im = sklearn.cluster.KMeans(n_clusters=5).fit(sel_feats[:,2:])
        
        # Create patch Image
        Patch_im = np.zeros((int(image_size[0]),int(image_size[1])))                        
        # Create superPatch Label image using the Superpatch size. 
        division_rows = np.floor(Patch_im.shape[0]/patch_size)
        division_cols = np.floor(Patch_im.shape[1]/patch_size)
        lins = np.repeat(list(range(int(division_cols))), patch_size)
        lins = np.repeat(np.expand_dims(lins,axis=1), patch_size, axis=1)
        for row_indx, row in enumerate(range(int(division_rows))):
            Patch_im[row_indx*patch_size:(row_indx+1)*patch_size,:int(division_cols*patch_size)] = np.transpose(lins+int(row_indx*division_cols))
        Patch_im = Patch_im.astype(int)

        # Cluster into distinct phenotypes
        Cluster_image_ALL = sel_clusters_all[Patch_im]
        Cluster_image = clusters_im.labels_[Patch_im]

        # Save cluster image
        imtosave = Image.fromarray(np.uint8(Cluster_image_ALL))
        imtosave.save(out_dir+name+'_cluster_ALL.png') 

        # Save cluster image
        imtosave = Image.fromarray(np.uint8(Cluster_image))
        imtosave.save(out_dir+name+'_cluster.png') 

def visualize_batch_effect_Ph(out_dir,features,features_idx,count):
    
    # Select n subjects randomly
    num_subjects = 5
    sel_subjects = sum(([features_idx==c for c in np.random.randint(0, high=features_idx.max(),size=num_subjects)]))>0
    features_idx = features_idx[sel_subjects]
    features = features[sel_subjects]
    
    # Select number of patches
    num_val = min(100000,features_idx.shape[0]*0.9)
    features_idx = features_idx[::int(features_idx.shape[0]/num_val)]
    features = features[::int(features.shape[0]/num_val),:]

    # Reorder samples
    sel_vals = np.random.randint(0,high=features.shape[0],size=features.shape[0])
    features = features[sel_vals]
    features_idx = features_idx[sel_vals]

    # Normalize features
    scaled_features = StandardScaler().fit_transform(features[:,2:])

    # Create UMAP
    embedding = TSNE(n_jobs=8,random_state=3, verbose=True).fit(scaled_features)

    # Create and save scatter plot figure
    plt.close('all')
    fig, ax = plt.subplots()   
    scatter = ax.scatter(embedding[:, 0],embedding[:, 1],s=0.5,c=features_idx,cmap='jet',edgecolors='none')
    legend1 = ax.legend(*scatter.legend_elements(),loc="upper right", title="Subjects")
    ax.add_artist(legend1)
    plt.title('TSNE of the dataset - one pat', fontsize=24)    
    plt.savefig(out_dir+'/TSNE_embeddings_Patients_'+str(count)+'.png',dpi=600)

    return

class SimCLR(object):

    def __init__(self, *args, **kwargs):
        self.args = kwargs['args']
        self.model = kwargs['model'].to(self.args['PCL_GPU_INDEX'])
        self.optimizer = kwargs['optimizer']
        self.scheduler = kwargs['scheduler']            
        self.model_dir = self.args['path']+'Patch_Contrastive_Learning/Model/'        
        self.writer = SummaryWriter(log_dir=self.model_dir)    
        if not os.path.exists(self.model_dir):
            os.makedirs(self.model_dir)  
        logging.basicConfig(filename=os.path.join(self.model_dir, 'training.log'), level=logging.DEBUG)
        self.criterion = torch.nn.CrossEntropyLoss().to(self.args['PCL_GPU_INDEX'])

    def info_nce_loss(self, features):

        labels = torch.cat([torch.arange(features.shape[0]/2) for i in range(2)], dim=0)
        labels = (labels.unsqueeze(0) == labels.unsqueeze(1)).float()
        labels = labels.to(self.args['PCL_GPU_INDEX'])

        features = F.normalize(features, dim=1)

        similarity_matrix = torch.matmul(features, features.T)

        # discard the main diagonal from both: labels and similarities matrix
        mask = torch.eye(labels.shape[0], dtype=torch.bool).to(self.args['PCL_GPU_INDEX'])
        labels = labels[~mask].view(labels.shape[0], -1)
        similarity_matrix = similarity_matrix[~mask].view(similarity_matrix.shape[0], -1)

        # select and combine multiple positives
        positives = similarity_matrix[labels.bool()].view(labels.shape[0], -1)

        # select only the negatives the negatives
        negatives = similarity_matrix[~labels.bool()].view(similarity_matrix.shape[0], -1)

        logits = torch.cat([positives, negatives], dim=1)
        labels = torch.zeros(logits.shape[0], dtype=torch.long).to(self.args['PCL_GPU_INDEX'])

        logits = logits / self.args['PCL_Temperature']
        return logits, labels

    def train(self, train_loader):
    
        scaler = GradScaler(enabled=True)

        # save config file
        save_config_file(self.writer.log_dir, self.args)

        # Train from previous checkpoint
        text_files = [f for f in os.listdir(self.model_dir) if f.endswith('.tar')]        
        if len(text_files)>0:
            text_files.sort()
            self.model.load_state_dict(torch.load(os.path.join(self.model_dir, text_files[-1])))
            self.model.train()
            done_epochs = len(text_files)*self.args['PCL_Epochs']/10
        else:
            done_epochs = 0

        n_iter = 0
        logging.info(f"Start SimCLR training for {self.args['PCL_Epochs']} epochs.")        

        for epoch_counter in tqdm(range(self.args['PCL_Epochs']),initial=int(done_epochs)):
            epoch_counter += int(done_epochs) 
            for images in train_loader:
                images[0] = torch.cat(torch.tensor_split(images[0], images[0].shape[0]),1).squeeze()
                images[1] = torch.cat(torch.tensor_split(images[1], images[1].shape[0]),1).squeeze()

                images = torch.cat((images[0],images[1]),dim=0)

                images = images.to(torch.float32)

                images = images.to(self.args['PCL_GPU_INDEX'])

                images = torch.moveaxis(images, 3, 1)

                with autocast(enabled=True):
                    features = self.model(images)
                    logits, labels = self.info_nce_loss(features)
                    loss = self.criterion(logits, labels)
                
                self.optimizer.zero_grad()

                scaler.scale(loss).backward()

                scaler.step(self.optimizer)
                scaler.update()

                # Log info
                top1, top5 = accuracy(logits, labels, topk=(1, 5))
                self.writer.add_scalar('loss', loss, global_step=n_iter)
                self.writer.add_scalar('acc/top1', top1[0], global_step=n_iter)
                self.writer.add_scalar('acc/top5', top5[0], global_step=n_iter)
                self.writer.add_scalar('learning_rate', self.scheduler.get_lr()[0], global_step=n_iter)

                n_iter += 1                
                
                del images, features
                gc.collect(0)
                gc.collect(1)
                gc.collect(2)
                logging.debug(f"Epoch: {epoch_counter}\tLoss: {loss}\tTop1 accuracy: {top1[0]}\tTop5 accuracy: {top5[0]}")

            # warmup for the first 10 epochs
            if epoch_counter >= 11*(self.args['PCL_Epochs']/10):
                self.scheduler.step()
            
            # save model checkpoints
            if epoch_counter % int(self.args['PCL_Epochs']/10) == 0:
                checkpoint_name = 'checkpoint_{:04d}.pth.tar'.format(epoch_counter)
                torch.save(self.model.state_dict(), os.path.join(self.model_dir, checkpoint_name))        
                logging.info(f"Model checkpoint and metadata has been saved at {self.model_dir}.")

            if epoch_counter>=self.args['PCL_Epochs']:
                break

        logging.info("Training has finished.")
        

    def infer(self,infer_loader, out_dir):
        
        # Open last saved model checkpoint          
        text_files = [f for f in os.listdir(self.model_dir) if f.endswith('.tar')]
        text_files.sort()
        self.model.load_state_dict(torch.load(os.path.join(self.model_dir, text_files[-1])))
        del self.model.backbone.fc._modules['4']
        del self.model.backbone.fc._modules['3']
        self.model.eval()       

        self.folder = out_dir 

        # Order Images 
        infer_loader.dataset.image_list.sort()

        # Extract patches and save it into a graph
        for patches, patch_pos, filename in tqdm(infer_loader):                                    
            
            # Locate model to device.
            self.model = self.model.to(self.args['PCL_GPU_INDEX'])
            
            # Minibatch to extract features
            step = self.args['PCL_N_Crops_per_Image']*self.args['PCL_Batch_Size']
            features = []
            for p in range(0,len(patches),step):                
                features.append(self.model(torch.moveaxis(torch.squeeze(torch.tensor(np.stack(patches[p:p+step],0),device=self.args['PCL_GPU_INDEX'])),3,1).to(torch.float32)).to('cpu').detach().numpy())
                gc.collect(0)
                gc.collect(1)      
                gc.collect(2)      
            features = np.concatenate(features,axis=0)

            # Save features
            features = np.concatenate((np.squeeze(patch_pos),features),axis=1)    

            # Join several images to one numpy structure
            patient_to_image = [i for i in os.listdir(self.args['path']+'Raw_Data/Experiment_Information/') if 'Patient_to_Image.xlsx'==i]
            if len(patient_to_image)>0:
                
                # Find subject name and subject file
                patient_to_image_excel = pd.read_excel(self.args['path']+'Raw_Data/Experiment_Information/'+patient_to_image[0])
                index = list(patient_to_image_excel['Image_Name']).index(filename[0])
                NPYname = self.folder+patient_to_image_excel['Subject_Name'][index]+'.npy'
                
                # if already exists...                 
                if os.path.isfile(NPYname):
                    
                    # Load previous concatenate with actual data and save it.
                    prevFeats = np.load(NPYname)
                    features[:,:2] = features[:,:2] + prevFeats[:,:2].max() + self.args['PCL_Patch_Size']*10 
                    np.save(NPYname,np.concatenate((prevFeats,features),axis=0))      
                    del prevFeats           

                else:

                    # Create new file
                    np.save(NPYname,features)    
                
            # Save one image is one subject
            else:
                np.save(self.folder+filename[0],features)    
                # tifffile.imsave(self.folder+filename[0]+'##.tiff',np.stack([np.reshape(features[:,i],(289,240)) for i in range(features.shape[1])]))


            # Eliminate data from garbage collector
            del patches, patch_pos, filename, features 
            gc.collect(0)
            gc.collect(1)      
            gc.collect(2)      
            
            
            # visualize_using_clustering(folder+filename[0],features,[1200,1200],patches.shape[2],image)

        return 


    def cluster_patches(self, patch_dir, out_dir):
        
        # Patches from subjects
        patch_images = os.listdir(patch_dir)        

        features = []
        features_idx = []

        # Extract patches and save it into a graph
        for subj_idx, filename in enumerate(patch_images):                                    
            
            # Load image 
            subject_feats = np.load(patch_dir+filename)

            # Store feats
            features.append(subject_feats)
            
            # Store subject idx
            features_idx.append([subj_idx]*subject_feats.shape[0])

        # Store in one file all features
        features = np.concatenate(features,axis=0)
        features_idx = np.concatenate(features_idx,axis=0)        

        # Show batch effect        
        visualize_batch_effect_Ph(out_dir,features,features_idx,0)

        # Cluster image 
        # visualize_using_clustering_ALL(self, out_dir, features, features_idx, patch_images)    

        return 